@@ -1,73 +1,103 @@
 package DTL::Fast::Template;
 use strict; use utf8; use warnings FATAL => 'all'; 
-use parent 'DTL::Fast::Template::Parser';
+use parent 'DTL::Fast::Parser';
 use Carp;
 
 our %TAG_HANDLERS;
 our %FILTER_HANDLERS;
 
 use DTL::Fast::Context;
-use DTL::Fast::Template::Tags;
-use DTL::Fast::Template::Filters;    
+use DTL::Fast::Tags;
+use DTL::Fast::Filters;    
 
 #@Override
 sub new
 {
-    my $proto = shift;
-    my $template = shift // '';
-    my $dirs = shift // []; # optional dirs to look up for includes or parents
-    my %kwargs = @_;
-    
-    $kwargs{'raw_chunks'} = _get_raw_chunks($template);
-    $kwargs{'dirs'} = $dirs;
-    $kwargs{'file_path'} //= 'inline';
-    $kwargs{'blocks'} = {};
-    
-    my $self = $proto->SUPER::new(%kwargs);
-  
-    return $self;
-}
+    my( $proto, $template, %kwargs ) = @_;
+    $template //= '';
 
-sub add_block
-{
-    my $self = shift;
-    my $block_name = shift;
-    my $block_ref = shift;
-    
-    if( exists $self->{'blocks'}->{$block_name} )
+    $kwargs{'raw_chunks'} = _get_raw_chunks($template);
+    $kwargs{'dirs'} //= [];             # optional dirs to look up for includes or parents
+    $kwargs{'file_path'} //= 'inline';  
+    $kwargs{'inherits'} //= {};
+    $kwargs{'inherited'} //= [];
+    $kwargs{'perl'} = $];
+    $kwargs{'modules'} //= {
+        'DTL::Fast' => $DTL::Fast::VERSION,
+    };
+
+    if( exists $kwargs{'inherits'}->{$kwargs{'file_path'}} )
     {
-        croak "Block $block_name is already registered. Duplicate names are not allowed";
+        croak  sprintf(
+            "Recursive inheritance detected:\n%s\n" 
+            , join "\n inherited from ", @{$kwargs{'inherited'}}, $kwargs{'file_path'}
+        );
     }
     
-    $self->{'blocks'}->{$block_name} = $block_ref;
+    $kwargs{'inherits'}->{$kwargs{'file_path'}} = $kwargs{'file_path'} eq 'inline' ? 
+        0: # this should depend on calling script
+        (stat($kwargs{'file_path'}))[9];
+        
+    push @{$kwargs{'inherited'}}, $kwargs{'file_path'};
     
-    return $self;
-}
+    my $self = $proto->SUPER::new(%kwargs);
 
-sub merge_blocks
-{
-    my $self = shift;
-    my $donor = shift;
-
-    foreach my $block_name ( keys %{$donor->{'blocks'}})
+    if( $self->{'extends'} )
     {
-        if( $self->{'blocks'}->{$block_name} )
+        my $parent_template = DTL::Fast::read_template(
+            $self->{'extends'}
+            , %kwargs
+        );
+          
+        if( defined $parent_template )
         {
-            $self->{'blocks'}->{$block_name}->{'chunks'} = $donor->{'blocks'}->{$block_name}->{'chunks'};
+            $self = $parent_template->merge_template($self);
+        }
+        else
+        {
+            croak  sprintf( "Couldn't found a parent template: %s in one of the following directories: %s"
+                , $self->{'extends' }
+                , join( ', ', @{$kwargs{'dirs'}})
+            );
         }
     }
     
     return $self;
 }
 
+sub merge_template
+{
+    my( $self, $donor ) = @_;
+
+    foreach my $block_name ( keys %{$donor->{'blocks'}})
+    {
+        if( my $block = $self->{'blocks'}->{$block_name} )
+        {
+            $block->replace_with($donor->{'blocks'}->{$block_name});
+        }
+    }
+
+    # appending module dependencies
+    my @modules = keys %{$donor->{'modules'}};
+    @{$self->{'modules'}}{@modules} = @{$donor->{'modules'}}{@modules};
+    
+    # appending inheritance path
+    @{$self}{'inherited','inherits'} = @{$donor}{'inherited','inherits'};
+    
+    return $self;
+}
+
+#@Override
+sub get_container_block{ return shift; }
+
 sub _get_raw_chunks
 {
-    my $template = shift;
+    my( $template ) = @_;
 
     my $reg = qr/(
-        \{\# .*? \#\}
-        |\{\% .+? \%\}
-        |\{\{ .+? \}\}
+        \{\#\s*.*?\s*\#\}
+        |\{\%\s*.+?\s*\%\}
+        |\{\{\s*.+?\s*\}\}
     )/x;
     my $result = [split /$reg/s, $template];
     
@@ -77,8 +107,7 @@ sub _get_raw_chunks
 #@Override
 sub render
 {
-    my $self = shift;
-    my $context = shift;
+    my( $self, $context ) = @_;
 
     $context //= {};
     
@@ -94,7 +123,7 @@ sub render
         croak  "Context must be a DTL::Fast::Context object or a HASH reference";
     }
     
-    $context->push();
+    $context->push_scope();
     
     $context->{'ns'}->[-1]->{'_dtl_ssi_dirs'} = $self->{'ssi_dirs'} if $self->{'ssi_dirs'};
     $context->{'ns'}->[-1]->{'_dtl_url_source'} = $self->{'url_source'} if $self->{'url_source'};
@@ -125,7 +154,7 @@ sub render
     pop @{$context->{'ns'}->[-1]->{'_dtl_include_path'}};
     delete $context->{'ns'}->[-1]->{'_dtl_include_files'}->{$template_path};
     
-    $context->pop();
+    $context->pop_scope();
     
     return $result;
 }
